@@ -116,21 +116,51 @@ void reportError(void *ptr, xmlErrorPtr error)
 {
     ErrorInfo *errorInfo = (ErrorInfo*) ptr;
 
-    if (errorInfo && errorInfo->verbose)
-    {
-        int line = (!errorInfo->filename)? 0 :
-            (errorInfo->xmlReader)?
-            xmlTextReaderGetParserLineNumber(errorInfo->xmlReader) :
-            error->line;
-        if (line)
-        {
-            fprintf(stderr, "%s:%d: ", errorInfo->filename, line);
-        }
-    }
-
     if (!errorInfo || errorInfo->verbose)
     {
+        int msglen;
+        int domain = error->domain;
+        const char *filename =
+            error->file? error->file :
+            errorInfo? errorInfo->filename :
+            NULL;
+        xmlTextReaderPtr reader = errorInfo? errorInfo->xmlReader : NULL;
+
+        int line = (!filename)? 0 :
+            (reader)? xmlTextReaderGetParserLineNumber(reader) :
+            error->line;
+        int column = (!filename)? 0 :
+            (reader)? xmlTextReaderGetParserColumnNumber(reader) :
+            error->int2;
+        if (line)
+        {
+            fprintf(stderr, "%s:%d.%d: ", filename, line, column);
+        }
+
+        msglen = strlen(error->message);
+        if (error->message[msglen-1] == '\n')
+            error->message[msglen-1] = '\0';
         fprintf(stderr, "%s", error->message);
+
+        /* only print extra info if it's not in message */
+        if (error->str1 && strstr(error->message, error->str1) == NULL) {
+            fprintf(stderr, ": %s", error->str1);
+            if (error->str2 && strstr(error->message, error->str2) == NULL) {
+                fprintf(stderr, ", %s", error->str2);
+            }
+            if (error->str3 && strstr(error->message, error->str3) == NULL) {
+                fprintf(stderr, ", %s", error->str3);
+            }
+        }
+        fprintf(stderr, "\n");
+
+
+        if ((domain == XML_FROM_PARSER) || (domain == XML_FROM_HTML) ||
+            (domain == XML_FROM_DTD) || (domain == XML_FROM_NAMESPACE) ||
+            (domain == XML_FROM_IO) || (domain == XML_FROM_VALID)) {
+            xmlParserCtxtPtr ctxt = error->ctxt;
+            if (ctxt) xmlParserPrintFileContext(ctxt->input);
+        }
     }
 }
 
@@ -240,3 +270,75 @@ main(int argc, char **argv)
     
     exit(ret);
 }
+
+static void bad_ns_opt(const char *msg)
+{
+    fprintf(stderr, "Bad namespace option: %s\n", msg);
+    exit(EXIT_BAD_ARGS);
+}
+
+#define MAX_NS_ARGS    256
+xmlChar *ns_arr[2 * MAX_NS_ARGS + 1];
+
+/**
+ *  Parse command line for -N <prefix>=<namespace> arguments
+ */
+int
+parseNSArr(xmlChar** ns_arr, int* plen, int argc, char **argv)
+{
+    int i = 0;
+    *plen = 0;
+    ns_arr[0] = 0;
+
+    for (i=0; i<argc; i++)
+    {
+        int prefix_len;
+        xmlChar *name, *value;
+        const xmlChar *equal_sign;
+
+        /* check for end of -N arguments */
+        if (argv[i] == 0 || argv[i][0] != '-' || strcmp(argv[i], "-N") != 0)
+            break;
+
+        i++;
+        if (i >= argc) bad_ns_opt("-N without argument");
+
+        equal_sign = xmlStrchr((const xmlChar*) argv[i], '=');
+        if (!equal_sign)
+            bad_ns_opt("namespace should have the form <prefix>=<url>");
+        prefix_len = equal_sign - (const xmlChar*) argv[i];
+
+        name = xmlStrndup((const xmlChar*) argv[i], prefix_len);
+        value = xmlStrdup((const xmlChar*) argv[i]+prefix_len+1);
+
+        if (*plen >= MAX_NS_ARGS)
+        {
+            fprintf(stderr, "too many namespaces increase MAX_NS_ARGS\n");
+            exit(EXIT_BAD_ARGS);
+        }
+
+        ns_arr[*plen] = name;
+        (*plen)++;
+        ns_arr[*plen] = value;
+        (*plen)++;
+        ns_arr[*plen] = 0;
+
+    }
+
+    return i;
+}
+
+/**
+ *  Cleanup memory allocated by namespaces arguments
+ */
+void
+cleanupNSArr(xmlChar **ns_arr)
+{
+    xmlChar **p = ns_arr;
+
+    while (*p)
+    {
+        xmlFree(*p);
+        p++;
+    }
+}
