@@ -34,7 +34,7 @@ THE SOFTWARE.
 
 #include <libxml/xmlmemory.h>
 #include <libxml/debugXML.h>
-#include <libxml/xmlIO.h>
+#include <libxml/xmlsave.h>
 #include <libxml/HTMLtree.h>
 #include <libxml/xinclude.h>
 #include <libxml/xpath.h>
@@ -49,20 +49,14 @@ THE SOFTWARE.
    TODO:
           1. Should this be allowed ?
              ./xml ed -m /xml /xml/table/rec/object ../examples/xml/tab-obj.xml 
-
-          2. Other options insert/append/update/rename/subnode (see tree.c)
-
-          3. Code clean-up is needed. (Too much code replication to find nodeset
-             by xpath expression).
 */
 
-#define MAX_NS_ARGS    256
-
 typedef struct _edOptions {   /* Global 'edit' options */
     int noblanks;             /* Remove insignificant spaces from XML tree */
     int preserveFormat;       /* Preserve original XML formatting */
     int omit_decl;            /* Omit XML declaration line <?xml version="1.0"?> */
     int inplace;              /* Edit file inplace (no output on stdout) */
+    int nonet;                /* Disallow network access */
 } edOptions;
 
 typedef edOptions *edOptionsPtr;
@@ -100,9 +94,6 @@ typedef struct _XmlEdAction {
 
 #define MAX_XML_ED_OPS 128
 
-static const char *ns_arr[2 * MAX_NS_ARGS + 1];
-int nCount = 0;
-
 int ops_count = 0;
 static XmlEdAction ops[MAX_XML_ED_OPS];
 
@@ -129,6 +120,7 @@ static const char edit_usage_str_2[] =
 "                        -N options must be last global options.\n";
 
 static const char edit_usage_str_3[] =
+"  --net               - allow network access"
 "  --help or -h        - display help\n\n"
 "where <action>\n"
 "  -d or --delete <xpath>\n"
@@ -140,15 +132,7 @@ static const char edit_usage_str_4[] =
 "  -m or --move <xpath1> <xpath2>\n"
 "  -r or --rename <xpath1> -v <new-name>\n"
 "  -u or --update <xpath> -v (--value) <value>\n"
-                  "\t\t\t -x (--expr) <xpath> (-x is not implemented yet)\n\n";
-
-/*
-   How to increment value of every attribute @weight?
-   How in --update refer to current value?
-   How to insert from a file?
-
-   xml ed --update "//elem/@weight" -x "./@weight+1"?
-*/
+"                         -x (--expr) <xpath>\n\n";
 
 /**
  *  display short help message
@@ -176,6 +160,7 @@ edInitOptions(edOptionsPtr ops)
     ops->omit_decl = 0;
     ops->preserveFormat = 0;
     ops->inplace = 0;
+    ops->nonet = 1;
 }
 
 /**
@@ -205,6 +190,10 @@ edParseOptions(edOptionsPtr ops, int argc, char **argv)
         {
             ops->inplace = 1;
         }
+        else if (!strcmp(argv[i], "--net"))
+        {
+            ops->nonet = 0;
+        }
         else if (!strcmp(argv[i], "--help") || !strcmp(argv[i], "-h") ||
                  !strcmp(argv[i], "-?") || !strcmp(argv[i], "-Z"))
         {
@@ -221,74 +210,29 @@ edParseOptions(edOptionsPtr ops, int argc, char **argv)
 }
 
 /**
- *  Parse command line for additional namespaces
+ *  register the namespace from @ns_arr to @ctxt
  */
-int
-edParseNSArr(const char** ns_arr, int* plen,
-              int count, char **argv)
+void
+nsarr_xpath_register(xmlXPathContextPtr ctxt)
 {
-    int i = 0;
-    *plen = 0;
-    ns_arr[0] = 0;
-
-    i = 0;
-    while(i < count)
-    {
-        if (argv[i] == 0) break;
-
-        /* printf("%d of %d %s\n", i, count, argv[i]); */
-
-        if (argv[i][0] == '-')
-        {
-            if (!strcmp(argv[i], "-N"))
-            {
-                int j;
-                xmlChar *name, *value;
-
-                i++;
-                if (i >= count) edUsage(argv[0], EXIT_BAD_ARGS);
-
-                for(j=0; argv[i][j] && (argv[i][j] != '='); j++);
-                if (argv[i][j] != '=') edUsage(argv[0], EXIT_BAD_ARGS);
-
-                name = xmlStrndup((const xmlChar *) argv[i], j);
-                value = xmlStrdup((const xmlChar *) argv[i]+j+1);
-
-                if (*plen >= MAX_NS_ARGS)
-                {
-                    fprintf(stderr, "too many namespaces increase MAX_NS_ARGS\n");
-                    exit(EXIT_BAD_ARGS);
-                }
-
-                ns_arr[*plen] = (char *)name;
-                (*plen)++;
-                ns_arr[*plen] = (char *)value;
-                (*plen)++;
-                ns_arr[*plen] = 0;
-
-                /* printf("xmlns:%s=\"%s\"\n", name, value); */
-            }
-        }
-        else
-            break;
-        i++;
+    int ns;
+    for (ns = 0; ns_arr[ns]; ns++) {
+        xmlXPathRegisterNs(ctxt, ns_arr[ns], ns_arr[ns+1]);
     }
-
-    return i;
 }
 
 /**
- *  Cleanup memory allocated by namespaces arguments
+ * register top-level namespace definitions from @doc to @ctxt
  */
-void
-edCleanupNSArr(const char **ns_arr)
+static void
+extract_ns_defs(xmlDocPtr doc, xmlXPathContextPtr ctxt)
 {
-    const char **p = ns_arr;
+    xmlNsPtr nsDef;
+    xmlNodePtr root = xmlDocGetRootElement(doc);
+    if (!root) return;
 
-    while (*p)
-    {
-        xmlFree((char *)*p);
-        p++;
+    for (nsDef = root->nsDef; nsDef; nsDef = nsDef->next) {
+        xmlXPathRegisterNs(ctxt, nsDef->prefix, nsDef->href);
     }
 }
 
@@ -296,561 +240,187 @@ edCleanupNSArr(const char **ns_arr)
  *  'update' operation
  */
 void
-edUpdate(xmlDocPtr doc, const char *loc, const char *val, XmlNodeType type)
+edUpdate(xmlDocPtr doc, xmlNodeSetPtr nodes, const char *val,
+    XmlNodeType type, xmlXPathContextPtr ctxt)
 {
-    int xptr = 0;
-    int expr = 0;
-    int tree = 0;
+    int i;
+    xmlXPathCompExprPtr xpath = NULL;
+    xmlNodePtr ctxt_node = ctxt->node;
 
-    xmlXPathObjectPtr res = NULL;
-    xmlXPathContextPtr ctxt;
-
-#if defined(LIBXML_XPTR_ENABLED)
-    if (xptr)
-    {
-        ctxt = xmlXPtrNewContext(doc, NULL, NULL);
-        res = xmlXPtrEval(BAD_CAST loc, ctxt);
+    if (type == XML_EXPR) {
+        xpath = xmlXPathCompile((const xmlChar*) val);
+        if (!xpath) return;
     }
-    else
+
+    for (i = 0; i < nodes->nodeNr; i++)
     {
-#endif
-        int ns = 0;
-        ctxt = xmlXPathNewContext(doc);
-        while(ns_arr[ns])
-        {
-            xmlXPathRegisterNs(ctxt, (const xmlChar *) ns_arr[ns], (const xmlChar *) ns_arr[ns+1]);
-            ns += 2;
+        /* update node */
+        if (type == XML_EXPR) {
+            xmlXPathObjectPtr res;
+            ctxt->node = nodes->nodeTab[i];
+            res = xmlXPathConvertString(xmlXPathCompiledEval(xpath, ctxt));
+            xmlNodeSetContent(nodes->nodeTab[i], res->stringval);
+            xmlXPathFreeObject(res);
+        } else {
+            /* TODO: do we need xmlEncodeEntitiesReentrant() too/instead? */
+            xmlChar *content = xmlEncodeSpecialChars(NULL, (const xmlChar*) val);
+            xmlNodeSetContent(nodes->nodeTab[i], content);
+            xmlFree(content);
         }
-
-        ctxt->node = xmlDocGetRootElement(doc);
-        if (expr)
-            res = xmlXPathEvalExpression(BAD_CAST loc, ctxt);
-        else
-        {
-            xmlXPathCompExprPtr comp;
-
-            comp = xmlXPathCompile(BAD_CAST loc);
-            if (comp != NULL)
-            {
-#if defined(LIBXML_DEBUG_ENABLED)
-                if (tree)
-                    xmlXPathDebugDumpCompExpr(stdout, comp, 0);
-#endif
-                res = xmlXPathCompiledEval(comp, ctxt);
-                xmlXPathFreeCompExpr(comp);
-            }
-            else res = NULL;
-        }
-#if defined(LIBXML_XPTR_ENABLED)
     }
-#endif
-    if (res == NULL) return;
 
-    /* Found loc */
-    switch(res->type)
-    {
-        case XPATH_NODESET:
-        {
-            int i;
-            xmlNodeSetPtr cur = res->nodesetval;
-            if (cur)
-            {
-                /*
-                fprintf(stderr, "Set contains %d nodes:\n", cur->nodeNr);
-                */
-                for (i = 0; i < cur->nodeNr; i++)
-                {
-                    /*
-                     *  update node
-                     */
-
-                    /* TODO: do we need xmlEncodeEntitiesReentrant() too/instead? */
-                    xmlChar *content = xmlEncodeSpecialChars(NULL, (const xmlChar*) val);
-
-                    xmlNodeSetContent(cur->nodeTab[i], content);
-                    xmlFree(content);
-                }
-            }
-            break;
-        }
-        default:
-            break;
-    }
-    xmlXPathFreeObject(res);
-    xmlXPathFreeContext(ctxt);
+    xmlXPathFreeCompExpr(xpath);
+    ctxt->node = ctxt_node;
 }
 
 /**
  *  'insert' operation
  */
 void
-edInsert(xmlDocPtr doc, const char *loc, const char *val, const char *name,
+edInsert(xmlDocPtr doc, xmlNodeSetPtr nodes, const char *val, const char *name,
          XmlNodeType type, int mode)
 {
-    int xptr = 0;
-    int expr = 0;
-    int tree = 0;
-
-    xmlXPathObjectPtr res = NULL;
-    xmlXPathContextPtr ctxt;
-
-#if defined(LIBXML_XPTR_ENABLED)
-    if (xptr)
+    int i;
+    for (i = 0; i < nodes->nodeNr; i++)
     {
-        ctxt = xmlXPtrNewContext(doc, NULL, NULL);
-        res = xmlXPtrEval(BAD_CAST loc, ctxt);
-    }
-    else
-    {
-#endif
-        int ns = 0;
-        ctxt = xmlXPathNewContext(doc);
-        while(ns_arr[ns])
+        /* update node */
+        if (type == XML_ATTR)
         {
-            xmlXPathRegisterNs(ctxt, (const xmlChar *) ns_arr[ns], (const xmlChar *) ns_arr[ns+1]);
-            ns += 2;
+            xmlNewProp(nodes->nodeTab[i], BAD_CAST name, BAD_CAST val);
         }
-
-        ctxt->node = xmlDocGetRootElement(doc);
-        if (expr)
-            res = xmlXPathEvalExpression(BAD_CAST loc, ctxt);
-        else
+        else if (type == XML_ELEM)
         {
-            xmlXPathCompExprPtr comp;
-
-            comp = xmlXPathCompile(BAD_CAST loc);
-            if (comp != NULL)
-            {
-#if defined(LIBXML_DEBUG_ENABLED)
-                if (tree)
-                    xmlXPathDebugDumpCompExpr(stdout, comp, 0);
-#endif
-                res = xmlXPathCompiledEval(comp, ctxt);
-                xmlXPathFreeCompExpr(comp);
-            }
-            else res = NULL;
+            xmlNodePtr node = xmlNewDocNode(doc, NULL /* TODO: NS */, BAD_CAST name, BAD_CAST val);
+            if (mode > 0)
+                xmlAddNextSibling(nodes->nodeTab[i], node);
+            else if (mode < 0)
+                xmlAddPrevSibling(nodes->nodeTab[i], node);
+            else
+                xmlAddChild(nodes->nodeTab[i], node);
         }
-#if defined(LIBXML_XPTR_ENABLED)
-    }
-#endif
-    if (res == NULL) return;
-
-    /* Found loc */
-    switch(res->type)
-    {
-        case XPATH_NODESET:
+        else if (type == XML_TEXT)
         {
-            int i;
-            xmlNodeSetPtr cur = res->nodesetval;
-            if (cur)
-            {
-                for (i = 0; i < cur->nodeNr; i++)
-                {
-                    /*
-                     *  update node
-                     */
-                    if (type == XML_ATTR)
-                    {
-                        xmlNewProp(cur->nodeTab[i], BAD_CAST name, BAD_CAST val);
-                    }
-                    else if (type == XML_ELEM)
-                    {
-                        xmlNodePtr node = xmlNewDocNode(doc, NULL /* TODO: NS */, BAD_CAST name, BAD_CAST val);
-                        if (mode > 0)
-                            xmlAddNextSibling(cur->nodeTab[i], node);
-                        else if (mode < 0)
-                            xmlAddPrevSibling(cur->nodeTab[i], node);
-                        else
-                            xmlAddChild(cur->nodeTab[i], node);
-                    }
-                    else if (type == XML_TEXT)
-                    {
-                        xmlNodePtr node = xmlNewDocText(doc, BAD_CAST val);
-                        if (mode > 0)
-                            xmlAddNextSibling(cur->nodeTab[i], node);
-                        else if (mode < 0)
-                            xmlAddPrevSibling(cur->nodeTab[i], node);
-                        else
-                            xmlAddChild(cur->nodeTab[i], node);                    
-                    }
-                }
-            }
-            break;
+            xmlNodePtr node = xmlNewDocText(doc, BAD_CAST val);
+            if (mode > 0)
+                xmlAddNextSibling(nodes->nodeTab[i], node);
+            else if (mode < 0)
+                xmlAddPrevSibling(nodes->nodeTab[i], node);
+            else
+                xmlAddChild(nodes->nodeTab[i], node);
         }
-        default:
-            break;
     }
-    xmlXPathFreeObject(res);
-    xmlXPathFreeContext(ctxt);
 }
 
 /**
  *  'rename' operation
  */
 void
-edRename(xmlDocPtr doc, char *loc, char *val, XmlNodeType type)
+edRename(xmlDocPtr doc, xmlNodeSetPtr nodes, char *val, XmlNodeType type)
 {
-    int xptr = 0;
-    int expr = 0;
-    int tree = 0;
-
-    xmlXPathObjectPtr res = NULL;
-    xmlXPathContextPtr ctxt;
-
-#if defined(LIBXML_XPTR_ENABLED)
-    if (xptr)
+    int i;
+    for (i = 0; i < nodes->nodeNr; i++)
     {
-        ctxt = xmlXPtrNewContext(doc, NULL, NULL);
-        res = xmlXPtrEval(BAD_CAST loc, ctxt);
+        xmlNodeSetName(nodes->nodeTab[i], BAD_CAST val);
     }
-    else
-    {
-#endif
-        int ns = 0;
-        ctxt = xmlXPathNewContext(doc);
-        while(ns_arr[ns])
-        {
-            xmlXPathRegisterNs(ctxt, (const xmlChar *) ns_arr[ns], (const xmlChar *) ns_arr[ns+1]);
-            ns += 2;
-        }
-
-        ctxt->node = xmlDocGetRootElement(doc);
-        if (expr)
-            res = xmlXPathEvalExpression(BAD_CAST loc, ctxt);
-        else
-        {
-            xmlXPathCompExprPtr comp;
-
-            comp = xmlXPathCompile(BAD_CAST loc);
-            if (comp != NULL)
-            {
-#if defined(LIBXML_DEBUG_ENABLED)
-                if (tree)
-                    xmlXPathDebugDumpCompExpr(stdout, comp, 0);
-#endif
-                res = xmlXPathCompiledEval(comp, ctxt);
-                xmlXPathFreeCompExpr(comp);
-            }
-            else res = NULL;
-        }
-#if defined(LIBXML_XPTR_ENABLED)
-    }
-#endif
-    if (res == NULL) return;
-
-    /* Found loc */    
-    switch(res->type)
-    {
-        case XPATH_NODESET:
-        {
-            int i;
-            xmlNodeSetPtr cur = res->nodesetval;
-            if (cur)
-            {
-                /*
-                fprintf(stderr, "Set contains %d nodes:\n", cur->nodeNr);
-                */
-                for (i = 0; i < cur->nodeNr; i++)
-                {
-                    xmlNodeSetName(cur->nodeTab[i], BAD_CAST val);
-                }
-            }
-            break;
-        }
-        default:
-            break;
-    }
-    xmlXPathFreeObject(res);
-    xmlXPathFreeContext(ctxt);
 }
 
 /**
  *  'delete' operation
  */
 void
-edDelete(xmlDocPtr doc, char *str)
+edDelete(xmlDocPtr doc, xmlNodeSetPtr nodes)
 {
-    int xptr = 0;
-    int expr = 0;
-    int tree = 0;
-
-    xmlXPathObjectPtr res = NULL;
-    xmlXPathContextPtr ctxt;
-
-#if defined(LIBXML_XPTR_ENABLED)
-    if (xptr)
+    int i;
+    for (i = nodes->nodeNr - 1; i >= 0; i--)
     {
-        ctxt = xmlXPtrNewContext(doc, NULL, NULL);
-        res = xmlXPtrEval(BAD_CAST str, ctxt);
-    }
-    else
-    {
-#endif
-        int ns = 0;
-        ctxt = xmlXPathNewContext(doc);
-        while(ns_arr[ns])
-        {
-            xmlXPathRegisterNs(ctxt, (const xmlChar *) ns_arr[ns], (const xmlChar *) ns_arr[ns+1]);
-            ns += 2;
+        if (nodes->nodeTab[i]->type == XML_NAMESPACE_DECL) {
+            fprintf(stderr, "FIXME: can't delete namespace nodes\n");
+            exit(EXIT_INTERNAL_ERROR);
         }
-        
-        ctxt->node = xmlDocGetRootElement(doc);
-        if (expr)
-            res = xmlXPathEvalExpression(BAD_CAST str, ctxt);
-        else
-        {
-            xmlXPathCompExprPtr comp;
+        /* delete node */
+        xmlUnlinkNode(nodes->nodeTab[i]);
 
-            comp = xmlXPathCompile(BAD_CAST str);
-            if (comp != NULL)
-            {
-#if defined(LIBXML_DEBUG_ENABLED)
-                if (tree)
-                    xmlXPathDebugDumpCompExpr(stdout, comp, 0);
-#endif
-                res = xmlXPathCompiledEval(comp, ctxt);
-                xmlXPathFreeCompExpr(comp);
-            }
-            else res = NULL;
-        }
-#if defined(LIBXML_XPTR_ENABLED)
+        /* Free node and children */
+        xmlFreeNode(nodes->nodeTab[i]);
+        nodes->nodeTab[i] = NULL;
     }
-#endif
-    if (res == NULL) return;
-
-    switch(res->type)
-    {
-        case XPATH_NODESET:
-        {
-            int i;
-            xmlNodeSetPtr cur = res->nodesetval;
-            if (cur)
-            {
-                /*
-                fprintf(stderr, "Set contains %d nodes:\n", cur->nodeNr);
-                */
-                /*for (i = 0; i < cur->nodeNr; i++)*/
-                for (i = cur->nodeNr - 1; i >= 0; i--)
-                {
-                    if (cur->nodeTab[i]->type == XML_NAMESPACE_DECL) {
-                        fprintf(stderr, "FIXME: can't delete namespace nodes\n");
-                        exit(EXIT_INTERNAL_ERROR);
-                    }
-                    /*
-                     *  delete node
-                     */
-                    xmlUnlinkNode(cur->nodeTab[i]);
-                    
-                    /*
-                     * Free node and children
-                     */
-                    xmlFreeNode(cur->nodeTab[i]);
-                    cur->nodeTab[i] = NULL;
-                }
-            }
-            
-            break;
-        }
-        default:
-            break;
-    }
-    xmlXPathFreeObject(res);
-    xmlXPathFreeContext(ctxt);
 }
 
 /**
  *  'move' operation
  */
 void
-edMove(xmlDocPtr doc, char *from, char *to)
+edMove(xmlDocPtr doc, xmlNodeSetPtr nodes, xmlNodePtr to)
 {
-    int xptr = 0;
-    int expr = 0;
-    int tree = 0;
-
-    xmlXPathObjectPtr res = NULL, res_to = NULL;
-    xmlXPathContextPtr ctxt;
-
-    /*
-     *  Find 'from' node set
-     */
-#if defined(LIBXML_XPTR_ENABLED)
-    if (xptr)
+    int i;
+    for (i = 0; i < nodes->nodeNr; i++)
     {
-        ctxt = xmlXPtrNewContext(doc, NULL, NULL);
-        res = xmlXPtrEval(BAD_CAST from, ctxt);
+        /* move node */
+        xmlUnlinkNode(nodes->nodeTab[i]);
+        xmlAddChild(to, nodes->nodeTab[i]);
     }
-    else
-    {
-#endif
-        int ns = 0;
-        ctxt = xmlXPathNewContext(doc);
-        while(ns_arr[ns])
-        {
-            xmlXPathRegisterNs(ctxt, (const xmlChar *) ns_arr[ns], (const xmlChar *) ns_arr[ns+1]);
-            ns += 2;
-        }
-
-        ctxt->node = xmlDocGetRootElement(doc);
-        if (expr)
-            res = xmlXPathEvalExpression(BAD_CAST from, ctxt);
-        else
-        {
-            xmlXPathCompExprPtr comp;
-
-            comp = xmlXPathCompile(BAD_CAST from);
-            if (comp != NULL)
-            {
-#if defined(LIBXML_DEBUG_ENABLED)
-                if (tree)
-                    xmlXPathDebugDumpCompExpr(stdout, comp, 0);
-#endif
-                res = xmlXPathCompiledEval(comp, ctxt);
-                xmlXPathFreeCompExpr(comp);
-            }
-            else res = NULL;
-        }
-#if defined(LIBXML_XPTR_ENABLED)
-    }
-#endif
-
-    if (ctxt) xmlXPathFreeContext(ctxt);
-    if (res == NULL) return;
-
-    /********************************************************/
-    
-    /*
-     *  Find 'to' node set
-     */
-#if defined(LIBXML_XPTR_ENABLED)
-    if (xptr)
-    {
-        ctxt = xmlXPtrNewContext(doc, NULL, NULL);
-        res_to = xmlXPtrEval(BAD_CAST to, ctxt);
-    }
-    else
-    {
-#endif
-        int ns = 0;
-        ctxt = xmlXPathNewContext(doc);
-        while(ns_arr[ns])
-        {
-            xmlXPathRegisterNs(ctxt, (const xmlChar *) ns_arr[ns], (const xmlChar *) ns_arr[ns+1]);
-            ns += 2;
-        }
-
-        ctxt->node = xmlDocGetRootElement(doc);
-        if (expr)
-            res_to = xmlXPathEvalExpression(BAD_CAST to, ctxt);
-        else
-        {
-            xmlXPathCompExprPtr comp;
-
-            comp = xmlXPathCompile(BAD_CAST to);
-            if (comp != NULL)
-            {
-#if defined(LIBXML_DEBUG_ENABLED)
-                if (tree)
-                    xmlXPathDebugDumpCompExpr(stdout, comp, 0);
-#endif
-                res_to = xmlXPathCompiledEval(comp, ctxt);
-                xmlXPathFreeCompExpr(comp);
-            }
-            else
-                res_to = NULL;
-        }
-#if defined(LIBXML_XPTR_ENABLED)
-    }
-#endif
-    if (res_to == NULL) return;
-
-    switch(res_to->type)
-    {
-        case XPATH_NODESET:
-        {
-            xmlNodeSetPtr cur = res_to->nodesetval;
-            if (cur)
-            {
-                if (cur->nodeNr != 1)
-                {
-                    fprintf(stderr, "destination nodeset does not contain one node (node count is %d)\n", cur->nodeNr);
-                    break;
-                }
-            }
-        }
-        default:
-            break;
-    }
-
-    switch(res->type)
-    {
-        case XPATH_NODESET:
-        {
-            int i;
-            xmlNodeSetPtr cur = res->nodesetval;
-            xmlNodeSetPtr cur_to = res_to->nodesetval;
-            if (cur && cur_to && (cur_to->nodeNr == 1))
-            {
-                for (i = 0; i < cur->nodeNr; i++)
-                {
-                    /*
-                     *  delete node
-                     */
-                    xmlUnlinkNode(cur->nodeTab[i]);
-                    xmlAddChild(res_to->nodesetval->nodeTab[0], cur->nodeTab[i]);             
-                }
-            }
-            break;
-        }
-        default:
-            break;
-    }
-
-    xmlXPathFreeObject(res);
-    xmlXPathFreeObject(res_to);
-    xmlXPathFreeContext(ctxt);
 }
 
 /**
  *  Loop through array of operations and perform them
  */
-int
+void
 edProcess(xmlDocPtr doc, XmlEdAction* ops, int ops_count)
 {
-    int res = 0;
     int k;
-    
+    xmlXPathContextPtr ctxt = xmlXPathNewContext(doc);
+    extract_ns_defs(doc, ctxt);
+    nsarr_xpath_register(ctxt);
+    ctxt->node = xmlDocGetRootElement(doc);
+
     for (k = 0; k < ops_count; k++)
     {
+        xmlXPathObjectPtr res;
+        xmlNodeSetPtr nodes;
+
+        res = xmlXPathEvalExpression(BAD_CAST ops[k].arg1, ctxt);
+        if (!res || res->type != XPATH_NODESET) continue;
+        nodes = res->nodesetval;
+
         switch (ops[k].op)
         {
             case XML_ED_DELETE:
-                edDelete(doc, ops[k].arg1);
+                edDelete(doc, nodes);
                 break;
-            case XML_ED_MOVE:
-                edMove(doc, ops[k].arg1, ops[k].arg2);
+            case XML_ED_MOVE: {
+                xmlXPathObjectPtr res_to =
+                    xmlXPathEvalExpression(BAD_CAST ops[k].arg2, ctxt);
+                if (!res_to
+                    || res_to->type != XPATH_NODESET
+                    || res_to->nodesetval->nodeNr != 1) {
+                    fprintf(stderr, "move destination is not a single node\n");
+                    continue;
+                }
+                edMove(doc, nodes, res_to->nodesetval->nodeTab[0]);
+                xmlXPathFreeObject(res_to);
                 break;
+            }
             case XML_ED_UPDATE:
-                edUpdate(doc, ops[k].arg1, ops[k].arg2, ops[k].type);
+                edUpdate(doc, nodes, ops[k].arg2, ops[k].type, ctxt);
                 break;
             case XML_ED_RENAME:
-                edRename(doc, ops[k].arg1, ops[k].arg2, ops[k].type);
+                edRename(doc, nodes, ops[k].arg2, ops[k].type);
                 break;
             case XML_ED_INSERT:
-                edInsert(doc, ops[k].arg1, ops[k].arg2, ops[k].arg3, ops[k].type, -1);
+                edInsert(doc, nodes, ops[k].arg2, ops[k].arg3, ops[k].type, -1);
                 break;
             case XML_ED_APPEND:
-                edInsert(doc, ops[k].arg1, ops[k].arg2, ops[k].arg3, ops[k].type, 1);
+                edInsert(doc, nodes, ops[k].arg2, ops[k].arg3, ops[k].type, 1);
                 break;
             case XML_ED_SUBNODE:
-                edInsert(doc, ops[k].arg1, ops[k].arg2, ops[k].arg3, ops[k].type, 0);
+                edInsert(doc, nodes, ops[k].arg2, ops[k].arg3, ops[k].type, 0);
                 break;
             default:
                 break;
         }
+        xmlXPathFreeObject(res);
     }
-
-    return res;
+    xmlXPathFreeContext(ctxt);
 }
 
 /**
@@ -859,12 +429,19 @@ edProcess(xmlDocPtr doc, XmlEdAction* ops, int ops_count)
 void
 edOutput(const char* filename, edOptions g_ops)
 {
-    xmlDocPtr doc = xmlReadFile(filename, NULL, 0);
-    int format = (g_ops.noblanks && !g_ops.preserveFormat);
+    xmlDocPtr doc;
+    int save_options =
+        (g_ops.noblanks? 0 : XML_SAVE_WSNONSIG) |
+        (g_ops.preserveFormat? 0 : XML_SAVE_FORMAT) |
+        (g_ops.omit_decl? XML_SAVE_NO_DECL : 0);
+    int read_options =
+        (g_ops.nonet? XML_PARSE_NONET : 0);
+    xmlSaveCtxtPtr save;
 
+    doc = xmlReadFile(filename, NULL, read_options);
     if (!doc)
     {
-        edCleanupNSArr(ns_arr);
+        cleanupNSArr(ns_arr);
         xmlCleanupParser();
         xmlCleanupGlobals();
         exit(EXIT_BAD_FILE);
@@ -872,32 +449,10 @@ edOutput(const char* filename, edOptions g_ops)
 
     edProcess(doc, ops, ops_count);
 
-    /* Print out result */
-    if (!g_ops.omit_decl)
-    {
-        xmlSaveFormatFile(g_ops.inplace ? filename : "-", doc, format);
-    }
-    else
-    {
-        int ret = 0;
-        char *encoding = NULL;
-        xmlOutputBufferPtr buf = NULL;
-        xmlCharEncodingHandlerPtr handler = NULL;
-        buf = xmlOutputBufferCreateFilename(g_ops.inplace ? filename : "-", handler, 0);
+    save = xmlSaveToFilename(g_ops.inplace? filename : "-", NULL, save_options);
+    xmlSaveDoc(save, doc);
+    xmlSaveClose(save);
 
-        if (doc->children != NULL)
-        {
-            xmlNodePtr child = doc->children;
-            while (child != NULL)
-            {
-                xmlNodeDumpOutput(buf, doc, child, 0, format, encoding);
-                xmlOutputBufferWriteString(buf, "\n");
-                child = child->next;
-            }
-        }
-        ret = xmlOutputBufferClose(buf);
-    }
-    
     xmlFreeDoc(doc);
 }
 
@@ -909,13 +464,14 @@ edMain(int argc, char **argv)
 {
     int i, j, n, start = 0;
     static edOptions g_ops;
+    int nCount = 0;
 
     if (argc < 3) edUsage(argv[0], EXIT_BAD_ARGS);
 
     edInitOptions(&g_ops);
     start = edParseOptions(&g_ops, argc, argv);
 
-    edParseNSArr(ns_arr, &nCount, argc-start, argv+start);
+    parseNSArr(ns_arr, &nCount, argc-start, argv+start);
         
     /*
      *  Parse command line and fill array of operations
@@ -1128,7 +684,7 @@ edMain(int argc, char **argv)
         edOutput(argv[n], g_ops);
     }
 
-    edCleanupNSArr(ns_arr);
+    cleanupNSArr(ns_arr);
     xmlCleanupParser();
     xmlCleanupGlobals();
     return 0;
