@@ -41,7 +41,6 @@ THE SOFTWARE.
 #include "xmlstar.h"
 #include "trans.h"
 
-#define MAX_NS_ARGS    256
 /* max length of xmlstarlet supplied (ie not from command line) namespaces
  * currently xalanredirect is longest, at 13 characters*/
 #define MAX_NS_PREFIX_LEN 20
@@ -78,6 +77,7 @@ lookup_ns_entry(const char *prefix, int len) {
 
 
 typedef struct _selOptions {
+    int quiet;            /* No output */
     int printXSLT;        /* Display prepared XSLT */
     int printRoot;        /* Print root element in output (if XML) */
     int outText;          /* Output is text */
@@ -122,15 +122,16 @@ static const char select_usage_str_1[] =
 
 static const char select_usage_str_2[] =
 "<global-options> are:\n"
+"  -Q or --quiet             - do not write anything to standard output.\n"
 "  -C or --comp              - display generated XSLT\n"
 "  -R or --root              - print root element <xsl-select>\n"
 "  -T or --text              - output is text (default is XML)\n"
 "  -I or --indent            - indent output\n"
 "  -D or --xml-decl          - do not omit xml declaration line\n"
-"  -B or --noblanks          - remove insignificant spaces from XML tree\n"
-"  -E or --encode <encoding> - output in the given encoding (utf-8, unicode...)\n";
+"  -B or --noblanks          - remove insignificant spaces from XML tree\n";
 
 static const char select_usage_str_3[] =
+"  -E or --encode <encoding> - output in the given encoding (utf-8, unicode...)\n"
 "  -N <name>=<value>         - predefine namespaces (name without \'xmlns:\')\n"
 "                              ex: xsql=urn:oracle-xsql\n"
 "                              Multiple -N options are allowed.\n"
@@ -144,11 +145,13 @@ static const char select_usage_str_4[] =
 "  -v or --value-of <xpath>  - print value of XPATH expression\n"
 "  -o or --output <string>   - output string literal\n"
 "  -n or --nl                - print new line\n"
-"  -f or --inp-name          - print input file name (or URL)\n";
+"  -f or --inp-name          - print input file name (or URL)\n"
+"  -m or --match <xpath>     - match XPATH expression\n";
 
 static const char select_usage_str_5[] =
-"  -m or --match <xpath>     - match XPATH expression\n"
 "  -i or --if <test-xpath>   - check condition <xsl:if test=\"test-xpath\">\n"
+"  --elif <test-xpath>       - check condition if previous conditions failed\n"
+"  --else                    - check if previous conditions failed\n"
 "  -e or --elem <name>       - print out element <xsl:element name=\"name\">\n"
 "  -a or --attr <name>       - add attribute <xsl:attribute name=\"name\">\n"
 "  -b or --break             - break nesting\n"
@@ -263,6 +266,7 @@ selUsage(const char *argv0, exit_status status)
 void
 selInitOptions(selOptionsPtr ops)
 {
+    ops->quiet = 0;
     ops->printXSLT = 0;
     ops->printRoot = 0;
     ops->outText = 0;
@@ -273,73 +277,6 @@ selInitOptions(selOptionsPtr ops)
     ops->encoding = NULL;
 }
 
-/**
- *  Parse command line for additional namespaces
- */
-int
-selParseNSArr(xmlChar** ns_arr, int* plen,
-              int count, char **argv)
-{
-    int i = 0;
-    *plen = 0;
-    ns_arr[0] = 0;
-
-    for (i=0; i<count; i++)
-    {
-        if (argv[i] == 0) break;
-        if (argv[i][0] == '-')
-        {
-            if (!strcmp(argv[i], "-N"))
-            {
-                int j;
-                xmlChar *name, *value;
-
-                i++;
-                if (i >= count) selUsage(argv[0], EXIT_BAD_ARGS);
-
-                for(j=0; argv[i][j] && (argv[i][j] != '='); j++);
-                if (argv[i][j] != '=') selUsage(argv[0], EXIT_BAD_ARGS);
-
-                name = xmlStrndup((const xmlChar *) argv[i], j);
-                value = xmlStrdup((const xmlChar *) argv[i]+j+1);
-
-                if (*plen >= MAX_NS_ARGS)
-                {
-                    fprintf(stderr, "too many namespaces increase MAX_NS_ARGS\n");
-                    exit(EXIT_BAD_ARGS);
-                }
-
-                ns_arr[*plen] = name;
-                (*plen)++;
-                ns_arr[*plen] = value;
-                (*plen)++;
-                ns_arr[*plen] = 0;
-
-                /*printf("xmlns:%s=\"%s\"\n", name, value);*/
-            }
-        }
-        else
-            break;
-    }
-
-    return i;
-}
-
-/**
- *  Cleanup memory allocated by namespaces arguments
- */
-void
-selCleanupNSArr(xmlChar **ns_arr)
-{
-    xmlChar **p = ns_arr;
-
-    while (*p)
-    {
-        xmlFree(*p);
-        p++;
-    }
-}
-
 /**
  *  Parse global command line options
  */
@@ -355,6 +292,10 @@ selParseOptions(selOptionsPtr ops, int argc, char **argv)
         {
             ops->printXSLT = 1;
         }
+        else if (!strcmp(argv[i], "-Q") || !strcmp(argv[i], "--quiet"))
+        {
+            ops->quiet = 1;
+        }
         else if (!strcmp(argv[i], "-B") || !strcmp(argv[i], "--noblanks"))
         {
             ops->noblanks = 1;
@@ -633,7 +574,7 @@ selPrepareXslt(xmlDocPtr style, selOptionsPtr ops, xmlChar *ns_arr[],
         xmlNewNs(root, ns_arr[ns+1], xmlStrlen(ns_arr[ns])?ns_arr[ns] : NULL);
         ns += 2;
     }
-    selCleanupNSArr(ns_arr);
+    cleanupNSArr(ns_arr);
 
     {
         xmlNodePtr output;
@@ -727,6 +668,67 @@ selPrepareXslt(xmlDocPtr style, selOptionsPtr ops, xmlChar *ns_arr[],
     return i;
 }
 
+/**
+ * copy top-level namespace definitions from @doc to @style_tree
+ */
+static void
+extract_ns_defs(xmlDocPtr doc, xmlDocPtr style_tree)
+{
+    xmlNsPtr nsDef;
+    xmlNodePtr style_root = xmlDocGetRootElement(style_tree);
+    xmlNodePtr root = xmlDocGetRootElement(doc);
+    if (!root) return;
+
+    for (nsDef = root->nsDef; nsDef; nsDef = nsDef->next) {
+        xmlNewNs(style_root, nsDef->href, nsDef->prefix);
+    }
+}
+
+static void
+do_file(const char *filename, xmlDocPtr style_tree,
+    int xml_options, const selOptions *ops, xsltOptions *xsltOps,
+    int *status)
+{
+    xmlChar *value;
+    xmlDocPtr doc;
+
+    /* Pass input file name as predefined parameter 'inputFile' */
+    const char *params[2+1] = { "inputFile" };
+    value = xmlStrdup((const xmlChar *)"'");
+    value = xmlStrcat(value, (const xmlChar *)filename);
+    value = xmlStrcat(value, (const xmlChar *)"'");
+    params[1] = (char *) value;
+
+
+    doc = xmlReadFile(filename, NULL, xml_options);
+    if (doc != NULL) {
+        xmlDocPtr res;
+
+        static xsltStylesheetPtr style = NULL;
+        if (!style) {
+            extract_ns_defs(doc, style_tree);
+            /* Parse XSLT stylesheet */
+            style = xsltParseStylesheetDoc(style_tree);
+            if (!style) exit(EXIT_LIB_ERROR);
+        }
+
+        res = xsltTransform(xsltOps, doc, params, style, filename);
+        if (!ops->quiet && res && xsltSaveResultToFile(stdout, res, style) < 0)
+        {
+            *status = EXIT_LIB_ERROR;
+        }
+        else if ((ops->quiet || *status == EXIT_FAILURE) && res->children)
+        {
+            *status = EXIT_SUCCESS;
+            if (ops->quiet) exit(EXIT_SUCCESS);
+        }
+    } else {
+        *status = EXIT_BAD_FILE;
+    }
+
+    xmlFree(value);
+}
+
 /**
  *  This is the main function for 'select' option
  */
@@ -735,26 +737,24 @@ selMain(int argc, char **argv)
 {
     static xsltOptions xsltOps;
     static selOptions ops;
-    static const char *params[2 * MAX_PARAMETERS + 1];
-    static xmlChar *ns_arr[2 * MAX_NS_ARGS + 1];
-    int start, i, n, status = 0;
+    int start, i, n, status = EXIT_FAILURE;
     int nCount = 0;
-    int nbparams;
     xmlDocPtr style_tree;
-    xsltStylesheetPtr style;
+    int xml_options = 0;
 
     if (argc <= 2) selUsage(argv[0], EXIT_BAD_ARGS);
 
     selInitOptions(&ops);
     xsltInitOptions(&xsltOps);
     start = selParseOptions(&ops, argc, argv);
+    xml_options |= ops.nonet? XML_PARSE_NONET : 0;
     xsltOps.nonet = ops.nonet;
     xsltOps.noblanks = ops.noblanks;
     xsltInitLibXml(&xsltOps);
     xsltSetSortFunc(caseSortFunction);
 
     /* set parameters */
-    selParseNSArr(ns_arr, &nCount, start, argv+2);
+    parseNSArr(ns_arr, &nCount, start, argv+2);
 
     style_tree = xmlNewDoc(NULL);
     i = selPrepareXslt(style_tree, &ops, ns_arr, start, argc, argv);
@@ -765,56 +765,15 @@ selMain(int argc, char **argv)
         exit(EXIT_SUCCESS);
     }
 
-    /*
-     *  Parse XSLT stylesheet
-     */
-    style = xsltParseStylesheetDoc(style_tree);
-    if (!style) exit(EXIT_LIB_ERROR);
-
     for (n=i; n<argc; n++)
-    {
-        xmlChar *value;
-
-        /*
-         *  Pass input file name as predefined parameter 'inputFile'
-         */
-        nbparams = 2;
-        params[0] = "inputFile";
-        value = xmlStrdup((const xmlChar *)"'");
-        value = xmlStrcat((xmlChar *)value, (const xmlChar *)argv[n]);
-        value = xmlStrcat((xmlChar *)value, (const xmlChar *)"'");
-        params[1] = (char *) value;
-
-        {
-            xmlDocPtr doc = xmlReadFile(argv[n], NULL, 0);
-            if (doc != NULL) {
-                xsltProcess(&xsltOps, doc, params, style, argv[n]);
-            } else {
-                status = 2;
-            }
-        }
-        xmlFree(value);
-    }
+        do_file(argv[n], style_tree, xml_options, &ops, &xsltOps, &status);
 
     if (i == argc)
-    {
-        xmlDocPtr doc;
-        nbparams = 2;
-        params[0] = "inputFile";
-        params[1] = "'-'";
-
-        doc = xmlReadFile("-", NULL, 0);
-        if (doc != NULL) {
-            xsltProcess(&xsltOps, doc, params, style, "-");
-        } else {
-            status = 2;
-        }
-    }
+        do_file("-", style_tree, xml_options, &ops, &xsltOps, &status);
 
     /* 
      * Shutdown libxml
      */
-    xsltFreeStylesheet(style);
     xsltCleanupGlobals();
     xmlCleanupParser();
     
